import asyncio
import inspect
import json
import os
from typing import List, Optional

from dosei_util import dosei_util
from croniter import croniter
from pydantic import BaseModel, field_validator


class FindDoseiInitError(Exception):
    pass


class CronJob(BaseModel):
    schedule: str
    entrypoint: str
    is_async: bool

    @field_validator("schedule")
    @classmethod
    def validate_schedule(cls, value):
        if croniter.is_valid(value) is False:
            raise ValueError("Invalid cron schedule format.")
        return value


class Dosei(BaseModel):
    _app_export_path: str = ".dosei/app.json"

    name: Optional[str] = None
    command: Optional[str] = None
    dev: Optional[str] = None
    port: Optional[int] = None
    cron_jobs: Optional[List[CronJob]] = []

    def cron_job(self, schedule: str):

        def decorator(func):
            self.cron_jobs.append(CronJob(
                schedule=schedule,
                entrypoint=f"{func.__module__}:{func.__name__}",
                is_async=Dosei.is_func_async(func)
            ))
            return func

        return decorator

    def export(self):
        os.makedirs(os.path.dirname(self._app_export_path), exist_ok=True)
        with open(self._app_export_path, 'w') as f:
            json.dump(self.model_dump(), f, indent=2)

    @staticmethod
    def is_func_async(func) -> bool:
        return inspect.iscoroutinefunction(func)

    @staticmethod
    def call_func(func):
        is_async = Dosei.is_func_async(func)
        return asyncio.run(func()) if is_async else func()

    @staticmethod
    def find_init(folder_path: str) -> str:
        return dosei_util.find_framework_init("Dosei", folder_path)



from fastapi import FastAPI


def get_fastapi_app(app):
    while hasattr(app, "app"):
        app = app.app
    return app


class FastAPIMiddleware:
    _self = None
    _initialized = False

    def __new__(cls, app):
        if cls._self is None:
            cls._self = super(FastAPIMiddleware, cls).__new__(cls)
        return cls._self

    def __init__(self, app):
        if self._initialized:
            return
        self.app = get_fastapi_app(app)
        self.deprecated_endpoints = [route.path for route in self.app.routes if getattr(route, 'deprecated', False)]
        self._initialized = True

    async def __call__(self, scope, receive, send):
        if scope['type'] == 'http':
            method = scope.get("method")
            path = scope.get("path")
            if path in self.deprecated_endpoints:
                print(f"Deprecated endpoint accessed: {method} {path}")

        await self.app(scope, receive, send)


def init(app: FastAPI):
    # TODO: This doesn't work, is breaking tests somehow, implement somewhere in the future
    # for app in gc.get_objects():
    #     if isinstance(app, FastAPI):
    #         app.add_middleware(FastAPIMiddleware)
    #         return
    app.add_middleware(FastAPIMiddleware)



import os
import subprocess
import sys

import click

from dosei import Dosei
from dosei.importer import import_from_string, ImportFromStringError

# Add current dir to PYTHONPATH
sys.path.insert(0, "")


@click.group()
def cli():
    pass


@cli.command()
@click.argument("func", required=False)
def run(func):
    if func:
        return Dosei.call_func(import_from_string(func))
    try:
        app: Dosei = import_from_string("dosei_config:dosei")
    except ImportFromStringError:
        raise click.ClickException(f"Couldn't find a dosei_config.py file in \"{os.getcwd()}\"")
    if app.dev is None:
        raise click.ClickException('Command "run" not found.')
    subprocess.check_call(app.command)


@cli.command()
def dev():
    try:
        app: Dosei = import_from_string("dosei_config:dosei")
    except ImportFromStringError:
        raise click.ClickException(f"Couldn't find a dosei_config.py file in \"{os.getcwd()}\"")
    if app.dev is None:
        raise click.ClickException('Command "dev" not found.')
    subprocess.check_call(app.dev)


@cli.command()
@click.argument("app")
def export(app):
    app: Dosei = import_from_string(app)
    return app.export()
